from steamship.app import Invocation, Request

from ..demo_apps.hello_world import HelloWorld, handler

NAME = "Ted"
RES_EMPTY = "Hello, Person"
RES_NAME = "Hello, {}".format(NAME)


def test_invoke_app_in_python():
    app = HelloWorld()

    assert (app.greet().body == RES_EMPTY)
    assert (app.greet(NAME).body == RES_NAME)


def test_invoke_app_with_request():
    app = HelloWorld()

    req = Request(invocation=Invocation(httpVerb="POST", appPath="greet"))
    res = app(req)
    assert (res.body == RES_EMPTY)

    req = Request(invocation=Invocation(httpVerb="POST", appPath="greet", arguments=dict(name=NAME)))
    res = app(req)
    assert (res.body == RES_NAME)


def test_invoke_app_with_handler():
    event = dict(
        invocation=dict(
            httpVerb="POST",
            appPath="greet"
        )
    )
    res = handler(event)
    assert (res['body'] == RES_EMPTY)

    event = dict(
        invocation=dict(
            httpVerb="POST",
            appPath="greet",
            arguments=dict(
                name=NAME
            )
        )
    )
    res = handler(event)
    assert (res['body'] == RES_NAME)



import json
import logging
from dataclasses import asdict
from typing import Any, TypeVar, Union, Dict

import requests  # type: ignore

from steamship.base.configuration import Configuration
from steamship.base.error import SteamshipError
from steamship.base.mime_types import MimeTypes
from steamship.base.request import Request
from steamship.base.response import Response, Task

__copyright__ = "Steamship"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

T = TypeVar('T', bound=Response)


class Client:
    """Client base.py class.

    Separated primarily as a hack to prevent circular imports.
    """

    # A client is always scoped by its space. A null space resolves to the
    # default space on the
    config: Configuration = None

    # Interaction prototype.
    dQuery: bool = False

    def __init__(
            self,
            apiKey: str = None,
            apiBase: str = None,
            appBase: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            profile: str = None,
            configFile: str = None,
            configDict: dict = None,
            dQuery: bool = False):

        self.config = Configuration(
            apiKey=apiKey,
            apiBase=apiBase,
            appBase=appBase,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            profile=profile,
            configFile=configFile,
            configDict=configDict
        )

        self.dQuery = dQuery

    def _url(
            self,
            appCall: bool = False,
            appOwner: str = None,
            operation: str = None,
            config: Configuration = None
    ):
        if not appCall:
            # Regular API call
            base = None
            if self.config and self.config.apiBase:
                base = self.config.apiBase
            if config and config.apiBase:
                base = config.apiBase
            if base is None:
                return SteamshipError(
                    code="EndpointMissing",
                    message="Can not invoke endpoint without the Client variable set.",
                    suggestion="This should automatically have a good default setting. Reach out to our Steamship support."
                )
        else:
            # Do the app version
            if appOwner is None:
                return SteamshipError(
                    code="UserMissing",
                    message="Can not invoke an app endpoint without the app owner's user handle.",
                    suggestion="Provide the appOwner option, or initialize your app with a lookup."
                )

            base = None
            if self.config and self.config.appBase:
                base = self.config.appBase
            if config and config.appBase:
                base = config.appBase
            if base is None:
                return SteamshipError(
                    code="EndpointMissing",
                    message="Can not invoke an app endpoint without the App Base variable set.",
                    suggestion="This should automatically have a good default setting. Reach out to our Steamship support."
                )
            if 'localhost' not in base and '127.0.0.1' not in base and '0:0:0:0' not in base:
                # We want to prepend the user handle
                parts = base.split('//')
                base = "{}//{}.{}".format(parts[0], appOwner, "//".join(parts[1:]))

        if base[len(base) - 1] == '/':
            base = base[:-1]
        if operation[0] == '/':
            operation = operation[1:]

        return "{}/{}".format(base, operation)

    def _headers(
            self,
            spaceId: str = None,
            spaceHandle: str = None,
            appCall: bool = False,
            appOwner: str = None,
            appId: str = None,
            appInstanceId: str = None
    ):
        ret = {
            "Authorization": "Bearer {}".format(self.config.apiKey)
        }

        sid = spaceId or self.config.spaceId
        shandle = spaceHandle or self.config.spaceHandle

        if sid:
            ret["X-Space-Id"] = sid
        elif shandle:
            ret["X-Space-Handle"] = shandle

        if appCall:
            if appOwner:
                ret['X-App-Owner-Handle'] = appOwner
            if appId:
                ret['X-App-Id'] = appId
            if appInstanceId:
                ret['X-App-Instance-Id'] = appInstanceId
        return ret

    def _data(
            self,
            verb: str,
            file: any,
            payload: Union[Request, dict]
    ):
        if type(payload) == dict:
            data = payload
        else:
            data = asdict(payload) if payload is not None else {}

        if verb == 'POST' and file is not None:
            # Note: requests seems to have a bug passing boolean (and maybe numeric?)
            # values in the midst of multipart form data. You need to manually convert
            # it to a string; otherwise it will pass as False or True (with the capital),
            # which is not standard notation outside of Python.
            for key in data:
                if data[key] is False:
                    data[key] = 'false'
                elif data[key] is True:
                    data[key] = 'true'

        return data

    def _response_data(self, resp, rawResponse: bool = False):
        if resp is None:
            return None

        if rawResponse:
            return resp.content

        if resp.headers:
            ct = None
            if 'Content-Type' in resp.headers:
                ct = resp.headers['Content-Type']
            if 'content-type' in resp.headers:
                ct = resp.headers['content-type']
            if ct is not None:
                ct = ct.split(';')[0]  # application/json; charset=utf-8
                if ct in [MimeTypes.TXT, MimeTypes.MKD, MimeTypes.HTML]:
                    return resp.text
                elif ct == MimeTypes.JSON:
                    return resp.json()
                else:
                    return resp.content

    def post(self, *args, **kwargs):
        return self.call('POST', *args, **kwargs)

    def get(self, *args, **kwargs):
        return self.call('GET', *args, **kwargs)

    def make_file_dict(self, data, file):
        result = {}
        for key, val in data.items():
            if val:
                if isinstance(val, Dict):
                    result[key] = (None, json.dumps(val), 'application/json')
                else:
                    result[key] = (None, str(val))
        result['file'] = file
        return result

    def call(
            self,
            verb: str,
            operation: str,
            payload: Union[Request, dict] = None,
            file: any = None,
            expect: any = None,
            asynchronous: bool = False,
            debug: bool = False,
            spaceId: str = None,
            spaceHandle: str = None,
            space: any = None,
            ifdQuery: bool = None,
            rawResponse: bool = False,
            appCall: bool = False,
            appOwner: str = None,
            appId: str = None,
            appInstanceId: str = None,
    ) -> Union[Any, Response[T]]:
        """Post to the Steamship API.

        All responses have the format:
           {
             data: <actual response>,
             error?: {
               reason: message
             }
           }

        For the Python client we return the contents of the `data`
        field if present, and we raise an exception if the `error`
        field is filled in.
        """
        if self.config.apiKey is None:
            raise Exception("Please set your Steamship API key.")

        if spaceId is None and space is not None and hasattr(space, 'id'):
            spaceId = getattr(space, 'id')

        if spaceId is None and spaceHandle is None and space is not None and hasattr(space, 'handle'):
            # Backup, if the spaceId transfer was None
            spaceHandle = getattr(space, 'handle')

        url = self._url(
            appCall=appCall,
            appOwner=appOwner,
            operation=operation
        )
        headers = self._headers(
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            appCall=appCall,
            appOwner=appOwner,
            appId=appId,
            appInstanceId=appInstanceId,
        )

        data = self._data(verb=verb, file=file, payload=payload)

        if verb == 'POST':
            if file is not None:
                files = self.make_file_dict(data, file)
                resp = requests.post(url, files=files, headers=headers)
            else:
                resp = requests.post(url, json=data, headers=headers)
        elif verb == 'GET':
            resp = requests.get(url, params=data, headers=headers)
        else:
            raise Exception("Unsupported verb: {}".format(verb))

        if debug is True:
            print("Response", resp)

        responseData = self._response_data(resp, rawResponse=rawResponse)

        if debug is True:
            print("Response JSON", responseData)

        task = None
        error = None
        obj = responseData

        if type(responseData) == dict:
            if 'status' in responseData:
                task = Task.from_dict(responseData['status'], client=self)
                # if task_resp is not None and task_resp.taskId is not None:
                #     task = Task(client=self)
                #     task.update(task_resp)
                if 'state' in responseData['status']:
                    if responseData['status']['state'] == 'failed':
                        error = SteamshipError.from_dict(responseData['status'], client=self)

            if 'data' in responseData:
                if expect is not None and hasattr(expect, 'from_dict'):
                    obj = expect.from_dict(responseData['data'], client=self)
                else:
                    obj = responseData['data']

            if 'reason' in responseData:
                # This is a legacy error reporting field. We should work toward being comfortable
                # removing this handler.
                error = SteamshipError(message=responseData['reason'])

        ret_obj = None
        if error is None:
            ret_obj = obj

        ret = Response[T](
            expect=expect,
            task=task,
            data=ret_obj,
            error=error,
            client=self
        )

        if ret.task is None and ret.data is None and ret.error is None:
            raise Exception('No data, task status, or error found in response')

        if self.dQuery is True and asynchronous:
            # This is an experimental UI for jQuery-style chaining.
            ret.wait()
            # In dQuery mode we throw an error to stop the chain
            if ret.error is not None:
                raise ret.error

        if self.dQuery is True and ifdQuery is not None:
            if ret.error is not None:
                raise ret.error
            return ifdQuery

        return ret



import json

from steamship import MimeTypes, DocTag
from steamship.base.response import TaskStatus
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag
from tests.client.helpers import _random_name, _steamship

__copyright__ = "Steamship"
__license__ = "MIT"


def test_file_upload():
    steamship = _steamship()
    name_a = "{}.mkd".format(_random_name())
    a = steamship.upload(
        name=name_a,
        content="A",
        mimeType=MimeTypes.MKD
    ).data
    assert (a.id is not None)
    assert (a.name == name_a)
    assert (a.mimeType == MimeTypes.MKD)

    name_b = "{}.txt".format(_random_name())
    b = steamship.upload(
        name=name_b,
        content="B",
        mimeType=MimeTypes.TXT
    ).data
    assert (b.id is not None)
    assert (b.name == name_b)
    assert (b.mimeType == MimeTypes.TXT)
    assert (a.id != b.id)

    name_c = "{}.txt".format(_random_name())
    c = steamship.upload(
        name=name_c,
        content="B",
        mimeType=MimeTypes.MKD
    ).data
    assert (c.mimeType == MimeTypes.MKD)  # The specified format gets precedence over filename

    d = steamship.upload(
        name=name_c,
        content="B",
    ).data
    assert (d.mimeType == MimeTypes.TXT)  # The filename is used in a pinch.

    a.delete()
    b.delete()
    c.delete()
    d.delete()


def test_file_scrape():
    steamship = _steamship()

    name_a = "{}.html".format(_random_name())
    a = steamship.scrape(
        name=name_a,
        url="https://edwardbenson.com/2020/10/gpt3-travel-agent"
    ).data
    assert (a.id is not None)
    assert (a.name == name_a)
    assert (a.mimeType == MimeTypes.HTML)

    name_b = "{}.html".format(_random_name())
    b = steamship.scrape(
        name=name_b,
        url="https://edwardbenson.com/2018/09/case-of-the-murderous-ai"
    ).data
    assert (b.id is not None)
    assert (a.id != b.id)
    assert (b.name == name_b)
    assert (b.mimeType == MimeTypes.HTML)

    a.delete()
    b.delete()


# def test_file_add_bloc():
#     steamship = _steamship()

#     name_a = "{}.txt".format(_random_name())
#     a = steamship.upload(
#         name=name_a,
#         content="This is a test."
#     )
#     assert(a.id is not None)
#     task  = a.convert()
#     task._run_development_mode()
#     task.wait()
#     q1 = a.query()
#     assert(len(q1.blocks) == 2)

#     # TODO: Append Content
#     # TODO: Append Blocks

def test_file_import_response_dict():
    resp = File.CreateResponse(bytes=b'some bytes', mimeType=MimeTypes.BINARY)
    to_dict = resp.to_dict()
    from_dict = File.CreateResponse.from_dict(to_dict)
    assert (resp.data == from_dict.data)
    assert (resp.mimeType == from_dict.mimeType)


def test_file_import_response_bytes_serialization():
    file_resp = File.CreateResponse(bytes=b'some bytes', mimeType=MimeTypes.BINARY)
    to_dict = file_resp.to_dict()
    as_json_string = json.dumps(to_dict)
    as_dict_again = json.loads(as_json_string)
    assert (as_dict_again == to_dict)


def test_file_upload_with_blocks():
    client = _steamship()
    name_a = "{}.mkd".format(_random_name())
    a = File.create(
        client=client,
        name=name_a,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(name="BlockTag")]),
            Block.CreateRequest(text="B")
        ],
        tags=[
            Tag.CreateRequest(name="FileTag")
        ]
    ).data
    assert (a.id is not None)
    assert (a.name == name_a)

    blocks = Block.listPublic(client, fileId=a.id)

    def check_blocks(blocks):
        assert (len(blocks) == 2)
        assert (blocks[0].tags is not None)
        assert (len(blocks[0].tags) == 1)
        assert (blocks[0].tags[0].name == "BlockTag")
        assert (blocks[0].text == "A")

    assert (blocks.data.blocks is not None)
    check_blocks(blocks.data.blocks)

    # Let's get the file fresh
    aa = File.get(client, id=a.id).data
    check_blocks(aa.blocks)
    assert (aa.tags is not None)
    assert (len(aa.tags) == 1)
    assert (aa.tags[0].name == "FileTag")

    a.delete()



from typing import Any
from dataclasses import dataclass
import logging
from typing import Union

@dataclass
class SteamshipError(Exception):
    message: str = None
    internalMessage: str = None
    suggestion: str = None
    code: str = None
    error: str = None

    def __init__(
            self,
            message: str = "Undefined remote error",
            internalMessage : str = None,
            suggestion: str = None,
            code: str = None,
            error: Union[Exception, str] = None):
        self.message = message
        self.suggestion = suggestion
        self.internalMessage = internalMessage
        self.statusCode = code
        if error is not None:
            self.error = "{}".format(error)

        parts = []
        if code is not None:
            parts.append("[{}]".format(code))
        if message is not None:
            parts.append(message)
        if internalMessage is not None:
            parts.append("Internal Message: {}".format(internalMessage))
        if suggestion is not None:
            parts.append("Suggestion: {}".format(suggestion))

        super().__init__("\n".join(parts))

    def log(self):
        logging.error("[{}] {}. [Internal: {}] [Suggestion: {}]".format(self.code, self.message, self.internalMessage, self.suggestion))
        if self.error:
            logging.error(self.error)

    def to_dict(self) -> dict:
        return dict(
            message=self.message,
            internalMessage=self.internalMessage,
            suggestion=self.suggestion,
            code=self.code,
            error=self.error
        )

    @staticmethod
    def from_dict(d: any, client: Any = None) -> "SteamshipError":
        """Last resort if subclass doesn't override: pass through."""
        return SteamshipError(
            message=d.get('statusMessage', d.get('message', None)),
            internalMessage=d.get('internalMessage', None),
            suggestion=d.get('statusSuggestion', suggestion=d.get('suggestion', None)),
            code=d.get('statusCode', d.get('code', None)),
            error=d.get('error', d.get('error', None))
        )



from dataclasses import dataclass

from steamship.base import Client, Request, Response
from steamship.base.request import GetRequest, IdentifierRequest


@dataclass
class Space:
    client: Client = None
    id: str = None
    name: str = None
    handle: str = None

    @dataclass
    class CreateRequest(Request):
        id: str = None
        name: str = None
        handle: str = None
        upsert: bool = None
        externalId: str = None
        externalType: str = None
        metadata: str = None

    @dataclass
    class ListRequest(Request):
        pass

    def delete(self) -> "Response[Space]":
        return self.client.post(
            'space/delete',
            IdentifierRequest(id=self.id),
            expect=Space
        )

    @staticmethod
    def from_dict(d: any, client: Client) -> "Space":
        if 'space' in d:
            d = d['space']

        return Space(
            client=client,
            id=d.get('id', None),
            name=d.get('name', None),
            handle=d.get('handle', None)
        )

    @staticmethod
    def get(
            client: Client,
            id: str = None,
            name: str = None,
            handle: str = None,
            upsert: bool = None,
            spaceId: str = None,
            spaceHandle: str = None,
            space: 'Space' = None
    ) -> "Response[Space]":
        req = GetRequest(
            id=id,
            name=name,
            handle=handle,
            upsert=upsert
        )
        return client.post('space/get', req, expect=Space, spaceId=spaceId, spaceHandle=spaceHandle, space=space)

    @staticmethod
    def create(
            client: Client,
            name: str,
            handle: str,
            externalId: str = None,
            externalType: str = None,
            metadata: any = None,
            upsert: bool = True
    ) -> "Response[Space]":
        req = Space.CreateRequest(
            name=name,
            handle=handle,
            upsert=upsert,
            externalId=externalId,
            externalType=externalType,
            metadata=metadata,
        )
        return client.post(
            'space/create',
            req,
            expect=Space
        )



from steamship import Block, File, DocTag, Tag
from steamship.data.tags import TagKind, DocTag
from steamship.app import App, post, create_handler, Response
from steamship.plugin.converter import Converter
from steamship.plugin.service import PluginResponse, PluginRequest
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

# Note 1: this aligns with the same document in the internal Engine test.
# Note 2: This should be duplicated from the test_importer because of the way our test system will
#         bundle this into an app deployment package (importing won't work!)
NAME = "Test Converter (Plugin)"
HANDLE = "test-converter-plugin-v1"
URL = "builtin://converter/test-plugin/v1"
TEST_H1 = "A Poem"
TEST_S1 = "Roses are red."
TEST_S2 = "Violets are blue."
TEST_S3 = "Sugar is sweet, and I love you."
TEST_DOC = "# {}\n\n{} {}\n\n{}\n".format(TEST_H1, TEST_S1, TEST_S2, TEST_S3)


class TestConverterPlugin(Converter, App):
    def run(self, request: PluginRequest[RawDataPluginInput]) -> PluginResponse[BlockAndTagPluginOutput]:
        return PluginResponse(data=BlockAndTagPluginOutput(file=File.CreateRequest(
            blocks=[
                Block.CreateRequest(text=TEST_H1, tags=[Tag.CreateRequest(kind=TagKind.doc, name=DocTag.h1)]),
                Block.CreateRequest(text=TEST_S1, tags=[Tag.CreateRequest(kind=TagKind.doc, name=DocTag.sentence)]),
                Block.CreateRequest(text=TEST_S2, tags=[Tag.CreateRequest(kind=TagKind.doc, name=DocTag.sentence)]),
                Block.CreateRequest(text=TEST_S3, tags=[Tag.CreateRequest(kind=TagKind.doc, name=DocTag.paragraph)])
            ]
        )))

    @post('convert')
    def convert(self, **kwargs) -> Response:
        rawDataPluginInput = Converter.parse_request(request=kwargs)
        blockAndTagPluginOutput = self.run(rawDataPluginInput)
        ret = Converter.response_to_dict(blockAndTagPluginOutput)
        return Response(json=ret)


handler = create_handler(TestConverterPlugin)



from steamship import MimeTypes, DocTag, PluginInstance

from tests.client.helpers import _random_index, _random_name, _steamship

__copyright__ = "Steamship"
__license__ = "MIT"

# TODO: It should fail if the docs field is empty.
# TODO: It should fail if the file hasn't been converted.

_TEST_EMBEDDER = "test-embedder"



def test_file_parse():
    steamship = _steamship()
    name_a = "{}.mkd".format(_random_name())
    T = "A nice poem"
    P1_1 = "Roses are red."
    P1_2 = "Violets are blue."
    P2_1 = "Sugar is sweet."
    P2_2 = "I love you."
    T2 = "A flavorful story"
    P3_1 = "Cake is made of flour."
    P3_2 = "Cake tastes good with milk."
    P4_1 = "Cake comes in chocolate and vanilla flavors."
    P4_2 = "Cake can be cut into many pieces and shared."

    content1 = "# {}\n\n{} {}\n\n{} {}".format(T, P1_1, P1_2, P2_1, P2_2)
    content2 = "# {}\n\n{} {}\n\n{} {}".format(T2, P3_1, P3_2, P4_1, P4_2)
    content = "{}\n\n{}".format(content1, content2)

    a = steamship.upload(
        name=name_a,
        content=content,
        mimeType=MimeTypes.MKD
    ).data
    assert (a.id is not None)
    assert (a.name == name_a)
    assert (a.mimeType == MimeTypes.MKD)

    convertResp = a.convert(pluginInstance="markdown-converter-default-1.0")
    assert (convertResp.error is None)
    convertResp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, pluginHandle='test-tagger').data
    parseResp = a.tag(pluginInstance=parser.handle)
    assert (parseResp.error is None)
    parseResp.wait()

    # Now the sentences should be parsed!
    q2 = a.query().data
    assert (len(q2.blocks) == 6)

    # Now we add the file to the index
    pluginInstance = PluginInstance.create(steamship, pluginHandle=_TEST_EMBEDDER).data
    with _random_index(steamship, pluginInstance=pluginInstance.handle) as index:
        index.insert_file(a.id, reindex=False)
        embedResp = index.embed()
        assert (embedResp.error is None)
        embedResp.wait()

        res = index.search("What color are roses?").data
        assert (len(res.items) == 1)
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert (res.items[0].value.value == " ".join([P1_1, P1_2]))

    a.delete()


def test_file_index():
    steamship = _steamship()
    name_a = "{}.mkd".format(_random_name())
    T = "A nice poem"
    P1_1 = "Roses are red."
    P1_2 = "Violets are blue."
    P2_1 = "Sugar is sweet."
    P2_2 = "I love you."
    T2 = "A flavorful story"
    P3_1 = "Cake is made of flour."
    P3_2 = "Cake tastes good with milk."
    P4_1 = "Cake comes in chocolate and vanilla flavors."
    P4_2 = "Cake can be cut into many pieces and shared."

    content1 = "# {}\n\n{} {}\n\n{} {}".format(T, P1_1, P1_2, P2_1, P2_2)
    content2 = "# {}\n\n{} {}\n\n{} {}".format(T2, P3_1, P3_2, P4_1, P4_2)
    content = "{}\n\n{}".format(content1, content2)

    a = steamship.upload(
        name=name_a,
        content=content,
        mimeType=MimeTypes.MKD
    ).data
    assert (a.id is not None)
    assert (a.name == name_a)
    assert (a.mimeType == MimeTypes.MKD)

    convertResp = a.convert(pluginInstance="markdown-converter-default-1.0")
    assert (convertResp.error is None)
    convertResp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, pluginHandle='test-tagger').data
    parseResp = a.tag(pluginInstance=parser.handle)
    assert (parseResp.error is None)
    parseResp.wait()

    # Now the sentences should be parsed!
    q2 = a.query().data
    assert (len(q2.blocks) == 6)

    # Now we add the file to the index via the shortcut.
    embedder = PluginInstance.create(steamship, pluginHandle='test-embedder').data
    index = a.index(pluginInstance=embedder.handle)

    res = index.search("What color are roses?").data
    assert (len(res.items) == 1)
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert (res.items[0].value.value == " ".join([P1_1, P1_2]))

    res = index.search("What flavors does cake come in?").data
    assert (len(res.items) == 1)
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert (res.items[0].value.value == " ".join([P4_1, P4_2]))

    index.delete()
    a.delete()


def test_file_embed_lookup():
    steamship = _steamship()
    name_a = "{}.mkd".format(_random_name())
    name_b = "{}.mkd".format(_random_name())

    content_a = "Ted likes to run."
    content_b = "Grace likes to bike."

    a = steamship.upload(
        name=name_a,
        content=content_a,
        mimeType=MimeTypes.MKD
    ).data

    convertRes = a.convert(pluginInstance="markdown-converter-default-1.0")
    assert (convertRes.error is None)
    convertRes.wait()

    parser = PluginInstance.create(steamship, pluginHandle='test-tagger').data
    parseRes = a.tag(pluginInstance=parser.handle)
    assert (parseRes.error is None)
    parseRes.wait()

    b = steamship.upload(
        name=name_b,
        content=content_b,
        mimeType=MimeTypes.MKD
    ).data
    convertRes = b.convert(pluginInstance="markdown-converter-default-1.0")
    assert (convertRes.error is None)
    convertRes.wait()

    parser = PluginInstance.create(steamship, pluginHandle='test-tagger').data
    parseRes = b.tag(pluginInstance=parser.handle)
    assert (parseRes.error is None)
    parseRes.wait()

    embedder = PluginInstance.create(steamship, pluginHandle='test-embedder').data
    # Now we add the file to the index
    with _random_index(steamship, embedder.handle) as index:
        index.insert_file(a.id, blockType='sentence', reindex=True)
        index.insert_file(b.id, blockType='sentence', reindex=True)

        res = index.search("What does Ted like to do?").data
        assert (len(res.items) == 1)
        assert (res.items[0].value.value == content_a)

        res = index.search("What does Grace like to do?").data
        assert (len(res.items) == 1)
        assert (res.items[0].value.value == content_b)

        # Now we list the items
        itemsa = index.list_items(fileId=a.id).data
        for item in itemsa.items:
            print("File {} - Value {}".format(item.fileId, item.value))
        assert (len(itemsa.items) == 1)
        assert (len(itemsa.items[0].embedding) > 0)
        assert (itemsa.items[0].value == content_a)

        itemsb = index.list_items(fileId=b.id).data
        assert (len(itemsb.items) == 1)
        assert (len(itemsb.items[0].embedding) > 0)
        assert (len(itemsb.items[0].embedding) == len(itemsa.items[0].embedding))
        assert (itemsb.items[0].value == content_b)



import logging
from typing import List

from steamship import Block
from steamship.base import Client, Response
from steamship.client.operations.embedder import EmbedRequest
from steamship.client.operations.tagger import TagRequest
from steamship.client.tasks import Tasks
from steamship.data import File
from steamship.data.embeddings import EmbedAndSearchRequest, QueryResults, EmbeddingIndex
from steamship.data.search import Hit
from steamship.data.space import Space

__copyright__ = "Steamship"
__license__ = "MIT"

from steamship.extension.file import TagResponse
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.embedded_items_plugin_output import EmbeddedItemsPluginOutput

_logger = logging.getLogger(__name__)



class Steamship(Client):
    """Steamship Python Client."""

    def __init__(
            self,
            apiKey: str = None,
            apiBase: str = None,
            appBase: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            profile: str = None,
            configFile: str = None,
            configDict: dict = None,
            dQuery: bool = False):
        super().__init__(
            apiKey=apiKey,
            apiBase=apiBase,
            appBase=appBase,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            profile=profile,
            configFile=configFile,
            configDict=configDict,
            dQuery=dQuery)
        """
        The base.py class will properly detect and set the defaults. They should be None here.
    
        dQuery is a Beta option that will return chainable responses, like:
          file.upload()
              .convert()
              .parse()
              .embed()
              .query()
    
        It offers no new functionality -- in fact at the moment it's slightly less in that you 
        are given the syntactically convenient response object for chaining rather than the actual
        response object of the invocation.
        """
        self.tasks = Tasks(self)

    def create_index(
            self,
            handle: str = None,
            name: str = None,
            pluginInstance: str = None,
            upsert: bool = True,
            externalId: str = None,
            externalType: str = None,
            metadata: any = None,
            spaceId: str = None,
            spaceHandle: str = None,
            space: Space = None
    ) -> Response[EmbeddingIndex]:
        return EmbeddingIndex.create(
            client=self,
            handle=handle,
            name=name,
            pluginInstance=pluginInstance,
            upsert=upsert,
            externalId=externalId,
            externalType=externalType,
            metadata=metadata,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        )

    def upload(
            self,
            filename: str = None,
            name: str = None,
            content: str = None,
            mimeType: str = None,
            plugin: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            space: Space = None
    ) -> File:
        return File.create(
            self,
            filename=filename,
            name=name,
            content=content,
            mimeType=mimeType,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        )

    def scrape(
            self,
            url: str,
            name: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            space: Space = None,
    ) -> File:
        if name is None:
            name = url
        return File.scrape(
            self,
            url,
            name,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        )

    def embed_and_search(
            self,
            query: str,
            docs: List[str],
            pluginInstance: str,
            k: int = 1,
            spaceId: str = None,
            spaceHandle: str = None,
            space: Space = None
    ) -> Response[QueryResults]:
        req = EmbedAndSearchRequest(
            query=query,
            docs=docs,
            pluginInstance=pluginInstance,
            k=k
        )
        return self.post(
            'plugin/instance/embeddingSearch',
            req,
            expect=QueryResults,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        )

    def tag(
            self,
            doc: str,
            pluginInstance: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            space: Space = None
    ) -> Response[TagResponse]:
        req = TagRequest(
            type="inline",
            file=File.CreateRequest(
                blocks=[Block.CreateRequest(
                    text=doc
                )]
            ),
            pluginInstance=pluginInstance,
        )
        return self.post(
            'plugin/instance/tag',
            req,
            expect=TagResponse,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        )



from dataclasses import dataclass

from steamship.base import Client


@dataclass
class User:
    client: Client = None
    id: str = None
    handle: str = None

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "User":
        if d['user'] is not None:
            d = d['user']

        return User(
            client=client,
            id=d.get("id", None),
            handle=d.get("handle", None),
        )

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            handle=self.handle
        )

    @staticmethod
    def current(client: Client) -> "User":
        return client.get(
            'account/current',
            expect=User
        )



from dataclasses import dataclass
from typing import Dict

from steamship.base import Client, Request


@dataclass
class CreateAppVersionRequest(Request):
    appId: str = None
    name: str = None
    handle: str = None
    upsert: bool = None
    type: str = 'file'
    configTemplate: Dict[str, any] = None


@dataclass
class DeleteAppVersionRequest(Request):
    id: str


@dataclass
class ListPrivateAppVersionsRequest(Request):
    pass


@dataclass
class AppVersion:
    client: Client = None
    id: str = None
    appId: str = None
    name: str = None
    handle: str = None
    configTemplate: Dict[str, any] = None

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "AppVersion":
        if 'appVersion' in d:
            d = d['appVersion']

        return AppVersion(
            client=client,
            id=d.get('id', None),
            name=d.get('name', None),
            handle=d.get('handle', None),
            configTemplate=d.get('configTemplate', None)
        )

    @staticmethod
    def create(
            client: Client,
            appId: str = None,
            name: str = None,
            handle: str = None,
            filename: str = None,
            filebytes: bytes = None,
            upsert: bool = None,
            configTemplate: Dict[str, any] = None
    ) -> "AppVersion":


        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, 'rb') as f:
                filebytes = f.read()

        req = CreateAppVersionRequest(
            name=name,
            handle=handle,
            appId=appId,
            upsert=upsert,
            configTemplate=configTemplate
        )

        return client.post(
            'app/version/create',
            payload=req,
            file=('app.zip', filebytes, "multipart/form-data"),
            expect=AppVersion
        )

    def delete(self) -> "AppVersion":
        req = DeleteAppVersionRequest(
            id=self.id
        )
        return self.client.post(
            'app/version/delete',
            payload=req,
            expect=AppVersion
        )



import io
import logging
from dataclasses import dataclass
from typing import Dict, Any

from steamship.base.binary_utils import flexi_create
from steamship.base import SteamshipError


@dataclass
class AppResponse:
    pass


@dataclass
class Http():
    status: int = None
    headers: Dict[str, str] = None


@dataclass
class Response(AppResponse):
    error: SteamshipError = None
    http: Http = None
    body: any = None

    def __init__(
            self,
            error: SteamshipError = None,
            http: Http = None,
            body: any = None,
            string: str = None,
            json: Any = None,
            bytes: io.BytesIO = None,
            mimeType=None
    ):
        if http is not None:
            self.http = http
        else:
            self.http = Http(status=200, headers={})
        self.error = error

        body, mimeType = flexi_create(
            body=body,
            string=string,
            json=json,
            bytes=bytes,
            mimeType=mimeType
        )

        self.body = body
        if mimeType is not None:
            self.http.headers["Content-Type"] = mimeType


def Error(
        httpStatus: int = 500,
        message: str = None,
        internalMessage: str = None,
        code: str = None,
        suggestion: str = None,
        error: Exception = None
) -> Response:
    err = SteamshipError(
        message=message,
        internalMessage=internalMessage,
        error=error,
        suggestion=suggestion,
        code=code
    )
    err.log()
    return Response(error=err, http=Http(status=httpStatus))



from abc import ABC
from dataclasses import dataclass
from typing import List

from steamship import File
from steamship.base import Client

@dataclass
class CorpusImportRequest:
    # The Corpus Identifiers
    client: Client = None
    id: str = None
    handle: str = None
    type: str = 'corpus'

    # Data for the plugin
    value: str = None
    data: str = None
    url: str = None
    pluginInstance: str = None
    fileImporterPluginInstance: str = None

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "CorpusImportRequest":
        return CorpusImportRequest(
            client=client,
            id=d.get('id', None),
            handle=d.get('handle', None),
            type='corpus',
            value=d.get('value', None),
            data=d.get('data', None),
            url=d.get('url', None),
            pluginInstance=d.get('pluginInstance', None),
            fileImporterPluginInstance=d.get('fileImporterPluginInstance', None)
        )

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            handle=self.handle,
            type=self.type,
            value=self.value,
            data=self.data,
            url=self.url,
            pluginInstance=self.pluginInstance,
            fileImporterPluginInstance=self.fileImporterPluginInstance
        )


@dataclass
class CorpusImportResponse:
    client: Client = None
    fileImportRequests: List[File.CreateRequest] = None

    def __init__(
            self,
            fileImportRequests: List[File.CreateRequest] = None
    ):
        self.fileImportRequests = fileImportRequests

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "CorpusImportResponse":
        return CorpusImportResponse(
            client=client,
            fileImportRequests=[File.CreateRequest.from_dict(req) for req in d.get('fileImportRequests', [])]
        )

    def to_dict(self) -> dict:
        return dict(
            fileImportRequests=self.fileImportRequests
        )



from steamship.base import Client, Response
from steamship.client.operations.tagger import TagRequest, TagResponse
from steamship.data.embeddings import EmbeddingIndex, EmbeddedItem
from steamship.data.file import File, FileUploadType
from steamship.data.plugin import PluginTargetType
from steamship.client.operations.converter import ConvertRequest
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput


@staticmethod
def upload(
        client: Client,
        filename: str = None,
        name: str = None,
        content: str = None,
        mimeType: str = None,
        corpusId: str = None,
        spaceId: str = None,
        spaceHandle: str = None,
        space: any = None
) -> "Response[File]":
    if filename is None and name is None and content is None:
        raise Exception("Either filename or name + content must be provided.")

    if filename is not None:
        with open(filename, 'rb') as f:
            content = f.read()
            name = filename

    req = File.CreateRequest(
        type=FileUploadType.file,
        corpusId=corpusId,
        name=name,
        mimeType=mimeType
    )

    return client.post(
        'file/create',
        payload=req,
        file=(name, content, "multipart/form-data"),
        expect=File,
        spaceId=spaceId,
        spaceHandle=spaceHandle,
        space=space
    )


File.upload = upload


def convert(self, pluginInstance: str = None ):
    req = ConvertRequest(
        type='file',
        id=self.id,
        pluginInstance=pluginInstance
    )

    return self.client.post(
        'plugin/instance/convert',
        payload=req,
        expect=BlockAndTagPluginOutput,
        asynchronous=True,
        ifdQuery=self
    )


File.convert = convert




def tag(
        self,
        pluginInstance: str = None,
        spaceId: str = None,
        spaceHandle: str = None,
        space: any = None
):
    req = TagRequest(
        type=PluginTargetType.file,
        id=self.id,
        pluginInstance=pluginInstance
    )

    return self.client.post(
        'plugin/instance/tag',
        payload=req,
        expect=TagResponse,
        asynchronous=True,
        ifdQuery=self,
        spaceId=spaceId,
        spaceHandle=spaceHandle,
        space=space
    )


File.tag = tag


# def tag(
#         self,
#         pluginInstance: str = None,
#         spaceId: str = None,
#         spaceHandle: str = None,
#         space: any = None
# ):
#     req = FileTagRequest(
#         id=self.id,
#         pluginInstance=pluginInstance
#     )
#
#     return self.client.post(
#         'file/tag',
#         payload=req,
#         expect=FileTagResponse,
#         asynchronous=True,
#         spaceId=spaceId,
#         spaceHandle=spaceHandle,
#         space=space
#     )

# File.tag = tag

def index(
        self,
        pluginInstance: str = None,
        indexName: str = None,
        indexId: str = None,
        index: "EmbeddingIndex" = None,
        upsert: bool = True,
        reindex: bool = True,
        spaceId: str = None,
        spaceHandle: str = None,
        space: any = None) -> "EmbeddingIndex":
    # TODO: This should really be done all on the app, but for now we'll do it in the client
    # to facilitate demos.

    if indexId is None and index is not None:
        indexId = index.id

    if indexName is None:
        indexName = "{}-{}".format(self.id, pluginInstance)

    if indexId is None and index is None:
        index = self.client.create_index(
            name=indexName,
            pluginInstance=pluginInstance,
            upsert=True,
            spaceId=spaceId,
            spaceHandle=spaceHandle,
            space=space
        ).data
    elif index is None:
        index = EmbeddingIndex(
            client=self.client,
            indexId=indexId
        )

    # We have an index available to us now. Perform the query.
    blocks = self.query(
        spaceId=spaceId,
        spaceHandle=spaceHandle,
        space=space
    ).data.blocks

    items = []
    for block in blocks:
        item = EmbeddedItem(
            value=block.text,
            externalId=block.id,
            externalType="block"
        )
        items.append(item)

    insert_task = index.insert_many(
        items,
        reindex=reindex,
        spaceId=spaceId,
        spaceHandle=spaceHandle,
        space=space
    )

    insert_task.wait()
    return index


File.index = index



from steamship.data.plugin import PluginType
from ..client.helpers import deploy_plugin, _steamship

__copyright__ = "Steamship"
__license__ = "MIT"

from tests.client.operations.test_embed import basic_embeddings, basic_embedding_search, count_embeddings
from tests.client.operations.test_embedding_index import create_index


def test_e2e_embedder():
    client = _steamship()
    with deploy_plugin("plugin_embedder.py", PluginType.tagger) as (plugin, version, instance):
        e1 = client.tag("This is a test", pluginInstance=instance.handle)
        e1.wait()
        assert (e1.error is None)
        assert (count_embeddings(e1.data.file) == 1)
        assert (len(e1.data.file.blocks[0].tags[0].value['embedding']) > 1)

        e2 = client.tag("This is a test", pluginInstance=instance.handle)
        e2.wait()
        assert (e2.error is None)
        assert (count_embeddings(e2.data.file) == 1)
        assert (len(e2.data.file.blocks[0].tags[0].value['embedding']) == len(e1.data.file.blocks[0].tags[0].value['embedding']))

        e4 = client.tag("This is a test", pluginInstance=instance.handle)
        e4.wait()
        assert (e4.error is None)
        assert (count_embeddings(e4.data.file) == 1)

        # Now lets run all the other embedding tests
        basic_embeddings(client, pluginInstance=instance.handle)
        basic_embedding_search(client, pluginInstance=instance.handle)

        create_index(client, pluginInstance=instance.handle)



import json

from steamship import Block, DocTag, Tag
from steamship.app import App, post, create_handler
from steamship.plugin.tagger import Tagger
from steamship.plugin.service import PluginResponse, PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
import re



class TestParserPlugin(Tagger, App):
    # For testing; mirrors TestConfigurableTagger in Swift

    def run(self, request: PluginRequest[BlockAndTagPluginInput]) -> PluginResponse[BlockAndTagPluginOutput]:
        tagKind = self.config['tagKind']
        tagName = self.config['tagName']
        tagValue = json.dumps({'numberValue':self.config['numberValue'], 'booleanValue':self.config['booleanValue']})

        if request.data is not None:
            file = request.data.file
            tag = Tag.CreateRequest(kind=tagKind, name= tagName, value=tagValue)
            if file.tags:
                file.tags.append(tag)
            else:
                file.tags = [tag]
            return PluginResponse(
                data=BlockAndTagPluginOutput(file)
            )

    @post('tag')
    def parse(self, **kwargs) -> dict:
        parseRequest = Tagger.parse_request(request=kwargs)
        parseResponse = self.run(parseRequest)
        return Tagger.response_to_dict(parseResponse)


handler = create_handler(TestParserPlugin)



#
# This is the CLIENT-side abstraction for an app.
#
# If you are implementing an app, see: steamship.app.server.App
#

from dataclasses import dataclass

from steamship.base import Client, Request


@dataclass
class CreateAppRequest(Request):
    id: str = None
    name: str = None
    handle: str = None
    upsert: bool = None


@dataclass
class DeleteAppRequest(Request):
    id: str


@dataclass
class ListPrivateAppsRequest(Request):
    pass


@dataclass
class App:
    client: Client = None
    id: str = None
    name: str = None
    handle: str = None

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "App":
        if 'app' in d:
            d = d['app']
        return App(
            client=client,
            id=d.get('id', None),
            name=d.get('name', None),
            handle=d.get('handle', None)
        )

    @staticmethod
    def create(
            client: Client,
            name: str = None,
            handle: str = None,
            upsert: bool = None
    ) -> "App":
        req = CreateAppRequest(
            name=name,
            handle=handle,
            upsert=upsert
        )
        return client.post(
            'app/create',
            payload=req,
            expect=App
        )

    def delete(self) -> "App":
        req = DeleteAppRequest(
            id=self.id
        )
        return self.client.post(
            'app/delete',
            payload=req,
            expect=App
        )



from dataclasses import dataclass
from typing import Dict

from steamship.base import Client
from steamship.data.file import File

@dataclass
class BlockAndTagPluginOutput():
    file: File.CreateRequest = None

    @staticmethod
    def from_dict(d: any = None, client: Client = None) -> "BlockAndTagPluginOutput":
        if d is None:
            return None

        return BlockAndTagPluginOutput(
            file=File.CreateRequest.from_dict(d.get('file', None), client=client)
        )

    def to_dict(self) -> Dict:
        if self.file is None:
            return dict()
        return dict(file=self.file.to_dict())



from dataclasses import dataclass
from typing import TypeVar, Any, List, Generic, Type

from steamship.base.base import IResponse
from steamship.base.metadata import str_to_metadata, metadata_to_str
from steamship.base.request import Request

T = TypeVar('T')


@dataclass
class CreateTaskCommentRequest(Request):
    taskId: str
    externalId: str = None
    externalType: str = None
    externalGroup: str = None
    metadata: str = None
    upsert: bool = None


@dataclass
class ListTaskCommentRequest(Request):
    taskId: str = None
    externalId: str = None
    externalType: str = None
    externalGroup: str = None


@dataclass
class DeleteTaskCommentRequest(Request):
    id: str = None


@dataclass
class TaskComment:
    client: any = None
    id: str = None
    userId: str = None
    taskId: str = None
    externalId: str = None
    externalType: str = None
    externalGroup: str = None
    metadata: any = None
    createdAt: str = None

    @staticmethod
    def create(
            client: Any,
            taskId: str = None,
            externalId: str = None,
            externalType: str = None,
            externalGroup: str = None,
            metadata: any = None,
            upsert: bool = True
    ) -> "IResponse[TaskComment]":
        req = CreateTaskCommentRequest(
            taskId=taskId,
            externalId=externalId,
            externalType=externalType,
            externalGroup=externalGroup,
            metadata=metadata_to_str(metadata),
            upsert=upsert
        )
        return client.post(
            'task/comment/create',
            req,
            expect=TaskComment,
        )

    @staticmethod
    def list(
            client: Any,
            taskId: str = None,
            externalId: str = None,
            externalType: str = None,
            externalGroup: str = None
    ) -> "IResponse[TaskCommentList]":
        req = ListTaskCommentRequest(
            taskId=taskId,
            externalId=externalId,
            externalType=externalType,
            externalGroup=externalGroup
        )
        return client.post(
            'task/comment/list',
            req,
            expect=TaskCommentList,
        )

    def delete(self) -> "IResponse[TaskComment]":
        req = DeleteTaskCommentRequest(self.id)
        return self.client.post(
            'task/comment/delete',
            req,
            expect=TaskComment,
        )

    @staticmethod
    def from_dict(d: any, client: Any = None) -> "TaskComment":
        return TaskComment(
            client=client,
            id=d.get('id', None),
            userId=d.get('userId', None),
            taskId=d.get('taskId', None),
            externalId=d.get('externalId', None),
            externalType=d.get('externalType', None),
            externalGroup=d.get('externalGroup', None),
            metadata=str_to_metadata(d.get("metadata", None)),
            createdAt=d.get('createdAt', None)
        )


@dataclass
class TaskCommentList:
    comments: List[TaskComment]

    @staticmethod
    def from_dict(d: any, client: Any = None) -> "TaskCommentList":
        return TaskCommentList(
            comments=[TaskComment.from_dict(dd, client) for dd in d.get('comments', [])]
        )


class TaskStatus:
    waiting = "waiting"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"


@dataclass
class TaskRunRequest(Request):
    taskId: str


@dataclass
class TaskStatusRequest(Request):
    taskId: str


@dataclass
class Task(Generic[T]):
    """Encapsulates a unit of asynchronously performed work."""
    client: Any = None
    taskId: str = None
    state: str = None
    statusMessage: str = None
    taskCreatedOn: str = None
    taskLastModifiedOn: str = None
    # eob: This is typed wrong, but I'm not sure how to type it..
    eventualResultType: Type[Any] = None

    @staticmethod
    def from_dict(d: any, client: Any = None) -> "Task":
        """Last resort if subclass doesn't override: pass through."""
        return Task(
            client=client,
            taskId=d.get('taskId', None),
            state=d.get('state', None),
            statusMessage=d.get('statusMessage', None),
            taskCreatedOn=d.get('taskCreatedOn', None),
            taskLastModifiedOn=d.get('taskLastModifiedOn', None)
        )

    def update(self, other: "Task"):
        """Incorporates a `Task` into this object."""
        if other is not None:
            self.taskId = other.taskId
            self.state = other.state
            self.statusMessage = other.statusMessage
            self.taskCreatedOn = other.taskCreatedOn
            self.taskLastModifiedOn = other.taskLastModifiedOn
        else:
            self.state = None

    def add_comment(self, externalId: str = None, externalType: str = None, externalGroup: str = None,
                    metadata: any = None, upsert: bool = True) -> IResponse[TaskComment]:
        return TaskComment.create(
            client=self.client,
            taskId=self.taskId,
            externalId=externalId,
            externalType=externalType,
            externalGroup=externalGroup,
            metadata=metadata,
            upsert=upsert
        )

    def list_comments(self) -> IResponse[TaskCommentList]:
        return TaskComment.list(client=self.client, taskId=self.taskId)

    def delete_comment(self, comment: TaskComment = None) -> IResponse[TaskComment]:
        return comment.delete()



import json
import os
from pathlib import Path
from typing import Dict

_configFile = '.steamship.json'


class Configuration:
    apiKey: str = None
    apiBase: str = None
    appBase: str = None
    webBase: str = None
    spaceId: str = None
    spaceHandle: str = None
    profile: str = None

    @staticmethod
    def from_dict(d: dict) -> "Configuration":
        if d is None:
            return Configuration()

        return Configuration(
            apiKey=d.get('apiKey', None),
            apiBase=d.get('apiBase', None),
            appBase=d.get('appBase', None),
            webBase=d.get('webBase', None),
            spaceId=d.get('spaceId', None),
            spaceHandle=d.get('spaceHandle', None)
        )

    def __init__(
            self,
            apiKey: str = None,
            apiBase: str = None,
            appBase: str = None,
            webBase: str = None,
            spaceId: str = None,
            spaceHandle: str = None,
            profile: str = None,
            configFile: str = None,
            configDict: dict = None,
    ):
        # First set the profile
        if "STEAMSHIP_PROFILE" in os.environ:
            self.profile = os.getenv('STEAMSHIP_PROFILE')
        if profile is not None:
            self.profile = profile

        # Then load from a file if provided
        if configFile is not None:
            self.load_from_file(configFile, self.profile)
        else:
            self.try_autofinding_files(self.profile)

        self.apply_env_var_overrides()

        if configDict is not None:
            self.merge_dict(configDict)

        self.apply_invocation_overrides(
            apiKey=apiKey,
            apiBase=apiBase,
            appBase=appBase,
            webBase=webBase,
            spaceId=spaceId,
            spaceHandle=spaceHandle
        )

        if self.apiBase is None:
            self.apiBase = 'https://api.steamship.com/api/v1/'
        if self.appBase is None:
            self.appBase = 'https://steamship.run/'
        if self.webBase is None:
            self.webBase = 'https://app.steamship.com/'

        if self.apiBase[len(self.apiBase) - 1] != '/':
            self.apiBase = "{}/".format(self.apiBase)

        if self.appBase[len(self.appBase) - 1] != '/':
            self.appBase = "{}/".format(self.appBase)

        if self.webBase[len(self.webBase) - 1] != '/':
            self.webBase = "{}/".format(self.webBase)

    def merge_dict(self, d: Dict[str, any]):
        apiKey = d.get('apiKey', None)
        if apiKey is not None:
            self.apiKey = apiKey

        apiBase = d.get('apiBase', None)
        if apiBase is not None:
            self.apiBase = apiBase

        appBase = d.get('appBase', None)
        if appBase is not None:
            self.appBase = appBase

        webBase = d.get('webBase', None)
        if webBase is not None:
            self.webBase = webBase

        profile = d.get('profile', None)
        if profile is not None:
            self.profile = profile

        spaceId = d.get('spaceId', None)
        if spaceId is not None:
            self.spaceId = spaceId

        spaceHandle = d.get('spaceHandle', None)
        if spaceHandle is not None:
            self.spaceHandle = spaceHandle

    def load_from_file(self, filepath: str, profile: str = None, throw_on_error=True):
        if not os.path.exists(filepath):
            if throw_on_error:
                raise Exception("Tried to load configuration file at {} but it did not exist.".format(filepath))
            else:
                return

        try:
            with open(filepath, 'r') as f:
                s = f.read()
                j = json.loads(s)

                if profile is None:
                    self.merge_dict(j)
                else:
                    if "profiles" not in j:
                        raise Exception("Profile {} requested but not found in {}".format(profile, filepath))
                    if profile not in j["profiles"]:
                        raise Exception("Profile {} requested but not found in {}".format(profile, filepath))
                    self.merge_dict(j["profiles"][profile])

        except Exception as err:
            if throw_on_error:
                raise err
            return

    def try_autofinding_files(self, profile: str = None):
        """
        Tries folders from cwd up to root.
        """
        paths = []
        cwd = Path(os.getcwd()).absolute()
        MAX_DEPTH = 40
        i = 0
        while len(str(cwd)) > 0 and str(cwd) != os.path.sep:
            paths.append(os.path.join(cwd, _configFile))
            cwd = cwd.parent.absolute()
            i += 1
            if i > MAX_DEPTH:
                print("ERROR: Max depth exceeded in config search recursion.")
                break

        paths.append(os.path.join(str(Path.home()), _configFile))
        for filepath in paths:
            if os.path.exists(filepath):
                self.load_from_file(filepath, profile=profile, throw_on_error=True)
                break  # Once we've found it; we're done.

    def apply_env_var_overrides(self):
        """Overrides with env vars"""
        if "STEAMSHIP_API_KEY" in os.environ:
            self.apiKey = os.getenv('STEAMSHIP_API_KEY')
        if "STEAMSHIP_API_BASE" in os.environ:
            self.apiBase = os.getenv('STEAMSHIP_API_BASE')
        if "STEAMSHIP_APP_BASE" in os.environ:
            self.appBase = os.getenv('STEAMSHIP_APP_BASE')
        if "STEAMSHIP_WEB_BASE" in os.environ:
            self.webBase = os.getenv('STEAMSHIP_WEB_BASE')
        if "STEAMSHIP_SPACE_ID" in os.environ:
            self.spaceId = os.getenv('STEAMSHIP_SPACE_ID')
        if "STEAMSHIP_SPACE_HANDLE" in os.environ:
            self.spaceHandle = os.getenv('STEAMSHIP_SPACE_HANDLE')

    def apply_invocation_overrides(
            self,
            apiKey: str = None,
            apiBase: str = None,
            appBase: str = None,
            webBase: str = None,
            spaceId: str = None,
            spaceHandle: str = None):
        if apiKey is not None:
            self.apiKey = apiKey
        if apiBase is not None:
            self.apiBase = apiBase
        if appBase is not None:
            self.appBase = appBase
        if webBase is not None:
            self.webBase = webBase
        if spaceId is not None:
            self.spaceId = spaceId
        if spaceHandle is not None:
            self.spaceHandle = spaceHandle



import json
from abc import ABC
from dataclasses import dataclass, asdict
from typing import List

from steamship.base import Client
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.service import PluginService, PluginRequest

# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Tagger.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.tagger instead
# of this file.
#
class Tagger(PluginService[BlockAndTagPluginInput, BlockAndTagPluginOutput], ABC):
    @classmethod
    def subclass_request_from_dict(cls, d: any, client: Client = None) -> PluginRequest[BlockAndTagPluginInput]:
        return BlockAndTagPluginInput.from_dict(d, client=client)

